/* 
   Support code to drive the USBMicro USB4xx I/O devices.

   Makes use of libusb to do USB specific pieces.

   Code isn't meant to be exhaustive and has only had basic testing
   but you should get the idea.  If you add any of the missing
   functions in please send me a patch :)

   Loosely based on sample code provided by USBMicro which we were
   told could be used any way we liked, including as part of a GPL'd
   package.

   (C) 2008,2009 Hugh Blemings, hugh@blemings.org

   Licensed under the GNU Public Licence V3

*/



#include <stdio.h>
#include <usb.h>

#include "usbmicro4xx.h"


/* Few static variables used by this code to talk to USB devices */
static struct usb_bus		*bus_list;
static struct usb_device	*dev = NULL;
static struct usb_dev_handle	*handle;
static int			busses, devices;

/* Scan the given usb_bus and find the first u4xx device
   support for multiple devices will need to change this
   function */
static struct usb_device *find_u4xx(struct usb_bus *bus) {

	struct usb_device *dev;
	/* Scan all buses, and every device within that bus */
	for(; bus; bus = bus->next) { 
		for(dev = bus->devices; dev; dev = dev->next) {
			if(dev->descriptor.idVendor == USBMICRO_VENDOR_ID &&
			   dev->descriptor.idProduct == USBMICRO_4XX_PRODUCT_ID) {
				return dev;
			}
		}
	}
	/* Not found */
	return NULL;
}

/* Send a command to usb4xx using USB control message, return data
   if any is put in command buffer */
static int send_command(struct usb_dev_handle *handle, char *command, 
		 int com_len, int res_len)
{
	int ret;

	ret = usb_control_msg(handle, 0x21, 9, 0x0200, 0, 
			      command, com_len, 5000);

	if(ret != com_len) {
		return USB4XX_CANT_SEND;
	}
	/* If we're expecting a result, process accordingly */
	if(res_len > 0) { 
		ret = usb_bulk_read(handle, 0x81, command, res_len, 5000);
		if(ret != res_len) {
			return USB4XX_CANT_READ;
		}
	}

	return ret;
}

/* Externally visible functions */

/* Initialise variables used by usb4xx code then open device 
   returns a handle to the initialised device to use in subsequent
   calls to I/O functions, or NULL on failure */
struct usb_dev_handle	*usb4xx_init(void)
{

	/* Call libusb to initialize the usb system */
	usb_init();

	handle = NULL;

	/* Populate statics we use internally in usb4xx code */
	busses = usb_find_busses();
	devices = usb_find_devices();
	bus_list = usb_get_busses();

	/* Look for the first u4xx device we can find then try and open */
	if((dev = find_u4xx(bus_list)) == NULL) {
		fprintf(stderr, "Couldn't find u4xx!\n");
		return NULL; 
	}
	if((handle = usb_open(dev)) == NULL) {
		fprintf(stderr, "Couldn't open u4xx!\n");
		return NULL;
	}

	/* The kernel's HID driver will seize the USBMicro device as it
	   says it's a HID device - we need to tell the kernel to 
	   let go of it */
	if (usb_detach_kernel_driver_np(handle, 0) < 0) {
	  /* If this fails, usually just means that no kernel driver
	     had attached itself to the device so just warn */
		fprintf(stderr, "Warning: Failed to detach kernel driver from USB device\n");
	}
	
	/* Set the configuration */
	if(usb_set_configuration(handle, 1) != 0) {
		perror("Couldn't set configuration for U4xx!");
		usb_close(handle);
		return NULL;
	}
  
	/* Clain interface */
	if (usb_claim_interface(handle, 0) != 0) {
		perror("Couldn't claim interface for U4xx!");
		usb_close(handle);
		return NULL;
	}

	return handle;
}

/* Close USB device */
void	usb4xx_close(usb_dev_handle *handle)
{
	
	usb_release_interface(handle, 0);
	usb_close(handle);

}

/* Commands to drive USB4xx device - these are based on the 
   descriptions in the USB Micro Documentation, the gentle reader
   is directed there for more information...

      http://www.usbmicro.com/odn/index.html

   ...in particular, refer to the "Raw Command Summary" section 
   under "Raw Device Programming" and the section on "DLL Commands"
   in the "USBm DLL Programming".
   
   The commands below should work identically to those in the
   USBm DLL with the exception that they expect a usb_dev_handle
   as the first argument rather than a integer to define the
   device. 

   If you want to support multiple devices some additional
   code will be required to identify them on the bus.

   Finally note that not all commands have been implemented,
   only those needed by the author for the particular task at 
   hand.  Contributions are of course welcome!
   
*/

int	usb4xx_init_ports(usb_dev_handle *handle) 
{
	int	ret;
	char	buffer[8] = {0, 0, 0, 0, 0, 0, 0, 0};

	ret = send_command(handle, buffer, 8, 0);
	if(ret != 8) {
		return ret;
	}
	return 0;
}

int usb4xx_write_a(usb_dev_handle *handle, u8 data)
{
	int	ret;
	char	buffer[8] = {0x01, 0, 0, 0, 0, 0, 0, 0};

	buffer[1] = (char) data;

	ret = send_command(handle, buffer, 8, 8);

	if(ret != 8) {
		return ret;
	}
	return 0;
}

int usb4xx_write_b(usb_dev_handle *handle, u8 data)
{
	int	ret;
	char	buffer[8] = {0x02, 0, 0, 0, 0, 0, 0, 0};

	buffer[1] = (char) data;
	ret = send_command(handle, buffer, 8, 8);

	if(ret != 8) {
		return ret;
	}
	return 0;
}

int usb4xx_write_a_bit(usb_dev_handle *handle, u8 and_term, u8 or_term)
{
	int	ret;
	char	buffer[8] = {0x03, 0, 0, 0, 0, 0, 0, 0};

	buffer[1] = (char) and_term;
	buffer[2] = (char) or_term;
	ret = send_command(handle, buffer, 8, 8);

	if(ret != 8) {
		return ret;
	}
	return 0;
}

int usb4xx_write_b_bit(usb_dev_handle *handle, u8 and_term, u8 or_term)
{
	int	ret;
	char	buffer[8] = {0x04, 0, 0, 0, 0, 0, 0, 0};

	buffer[1] = (char) and_term;
	buffer[2] = (char) or_term;
	ret = send_command(handle, buffer, 8, 8);

	if(ret != 8) {
		return ret;
	}
	return 0;
}

int	usb4xx_read_a(usb_dev_handle *handle) 
{
	int	ret;

	char	buffer[8] = {0x05, 0, 0, 0, 0, 0, 0, 0};

	ret = send_command(handle, buffer, 8, 8);

	if(ret != 8) {
		return ret;
	}

	return (buffer[1] & 0xff);
}

int	usb4xx_read_b(usb_dev_handle *handle) 
{
	int	ret;

	char	buffer[8] = {0x06, 0, 0, 0, 0, 0, 0, 0};

	ret = send_command(handle, buffer, 8, 8);

	if(ret != 8) {
		return ret;
	}

	return (buffer[1] & 0xff);
}

/* Missing: SetBit - 0x07 */
/* Missing: ResetBit - 0x08 */

int	usb4xx_direction_a_bit(usb_dev_handle *handle, u8 dir0, u8 dir1)
{
	int	ret;
	char	buffer[8] = {0x09, 0, 0, 0, 0, 0, 0, 0};

	buffer[1] = (char) dir0;
	buffer[2] = (char) dir1;
	ret = send_command(handle, buffer, 8, 8);

	if(ret != 8) {
		return ret;
	}
	return 0;
}

int	usb4xx_direction_b_bit(usb_dev_handle *handle, u8 dir0, u8 dir1)
{
	int	ret;
	char	buffer[8] = {0x0a, 0, 0, 0, 0, 0, 0, 0};

	buffer[1] = (char) dir0;
	buffer[2] = (char) dir1;
	ret = send_command(handle, buffer, 8, 8);

	if(ret != 8) {
		return ret;
	}
	return 0;
}

/* Number of functions need to be implemented here... */

/* Missing: Strobe Write - 0x0b */
/* Missing: Strobe Read - 0x0c */
/* Missing: Multi-Byte StrobeWrite - 0x0d */
/* Missing: Multi-Byte StrobeWrite - 0x0e */
/* Unused: 0x0f */
/* Missing: Init LCD - 0x10 */
/* Missing: Write LCD Command- 0x11 */
/* Missing: Write LCD Data - 0x12 */
/* Unused: 0x13 */

int	usb4xx_init_spi(usb_dev_handle *handle, u8 data)
{
	int	ret;
	char	buffer[8] = {0x0a, 0, 0, 0, 0, 0, 0, 0};

	buffer[1] = (char) data;
	ret = send_command(handle, buffer, 8, 8);

	if(ret != 8) {
		return ret;
	}
	return 0;
}

int	usb4xx_spi_master(usb_dev_handle *handle, int num_bytes, u8 *data)
{
}

int	usb4xx_spi_slave_write(usb_dev_handle *handle, int index, u8 *data)
{
}

int	usb4xx_spi_slave_read(usb_dev_handle *handle, int *count, u8 *data)
{
}

